from abc import abstractmethod
from dataclasses import dataclass

import polars as pl
from textual.message import Message
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import Label

COLORS = pl.Enum(
    (
        "#576176",
        "#FAA5AB",
        "#A5CD84",
        "#EFBD58",
        "#8DC3F1",
        "#DEAEED",
        "#27FFDF",
        "#CACCD3",
    )
)

COLORS_STYLES = "\n".join(
    f"""
ExtendedDataTable > .datatable--row{i} {{
    color: {x};
}}
"""
    for i, x in enumerate(COLORS.categories)
)

INDEX_COL = "__index"
COLOR_COL = "__color"
DISPLAY_IDX_COL = "_display_index"


class HasState:

    @abstractmethod
    def save_state(self, existing: dict) -> dict:
        """
        Generate any persistent data from this object

        Args:
            existing: Any existing state for this object which should be merged with the current state.
            (e.g if there are multiple instances of the browser which should be merged into a single state object)
        """

    @abstractmethod
    def load_state(self, state: dict, table_name: str, df: pl.DataFrame):
        """
        Apply the provided state to the current object

        Args:
            state: the state
            table_name: the current table name being displayed
            df: The full dataframe being displayed
        """


class ReactiveLabel(Label):

    value: reactive[str] = reactive("", layout=True)

    def render(self):
        if isinstance(self.value, (int, float)):
            return f"{self.value:,}"
        return str(self.value)


class ReceivesTableSelect(Widget):

    BINDINGS = [
        ("ctrl+t", "select_from_table()", "Select/copy value from table"),
    ]

    def action_select_from_table(self):
        self.post_message(SelectFromTable(interested_widget=self))

    @abstractmethod
    def on_table_select(self, value: str):
        pass


@dataclass
class SelectFromTable(Message):
    interested_widget: ReceivesTableSelect



from dataclasses import dataclass

import polars as pl
from textual import on
from textual.message import Message
from textual.widget import Widget

from dt_browser import INDEX_COL
from dt_browser.custom_table import CustomTable


def remove_row(dt: pl.DataFrame, rem_row: int):
    above = dt.slice(0, rem_row)
    below = dt.slice(rem_row + 1)
    return pl.concat([above, below])


class Bookmarks(Widget):

    @dataclass
    class BookmarkSelected(Message):
        selected_index: int

    @dataclass
    class BookmarkRemoved(Message):
        selected_index: int

    DEFAULT_CSS = """
    Bookmarks {
        dock: bottom;
        height: 15;
        border: tall white;

    }

    .bookmarks--history {
        padding: 0 1;
    }
    """

    BINDINGS = [
        ("escape", "close()", "Close"),
        ("delete", "remove_bookmark()", "Remove bookmark"),
    ]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._bookmark_df = pl.DataFrame()
        self._history: dict[str, list[int]] = {}
        self._meta_dt = pl.DataFrame()

    @property
    def meta_dt(self):
        return self._meta_dt

    # def save_state(self, existing: dict):
    #     max_history = 10
    #     history = self._history.copy()
    #     for k, v in existing[history]:
    #         if k not in history:
    #             history[k] = v

    #     if len(history) > max_history:
    #         history = {k: v for k, v in list(history.items())[-max_history:]}
    #     return {"history": {**history, self._table_name: self._bookmark_df.data[_INDEX_COL].to_dict()}}

    # def load_state(self, state: dict, table_name: str, df: pl.DataFrame):
    #     self._history = state["history"]
    #     if self._table_name in self._history:
    #         self.add_bookmark(df[*state["history"][table_name]])

    def compose(self):
        dt = CustomTable(self._bookmark_df, metadata_dt=self._meta_dt, cursor_type=CustomTable.CursorType.ROW)
        dt.styles.height = "auto"
        yield dt

    def toggle_bookmark(self, df: pl.DataFrame, meta_df: pl.DataFrame):
        if not self._meta_dt.is_empty() and meta_df[INDEX_COL][0] in self._meta_dt[INDEX_COL]:
            self.remove_bookmark(meta_df[INDEX_COL][0])
            return False
        self._bookmark_df = pl.concat([self._bookmark_df, df])
        self._meta_dt = pl.concat([self._meta_dt, meta_df])
        dt = self.query_children(CustomTable)
        if dt:
            dt.first().set_dt(self._bookmark_df, self._meta_dt)
        return True

    @property
    def has_bookmarks(self):
        return not self._bookmark_df.is_empty()

    def action_close(self):
        self.remove()

    async def on_mount(self):
        self.query_one(CustomTable).focus()

    def action_remove_bookmark(self):
        dt = self.query_one(CustomTable)
        idx = self._meta_dt[dt.cursor_coordinate.row][INDEX_COL][0]
        self.remove_bookmark(idx)

    def remove_bookmark(self, idx: int):
        dt = self.query_children(CustomTable)
        if len(self._bookmark_df) == 1:
            self._bookmark_df = self._bookmark_df.clear()
            self._meta_dt = self._meta_dt.clear()
            if dt:
                self.remove()
        else:
            rem_row = self._meta_dt.with_row_index(name="__tmp").filter(pl.col(INDEX_COL) == idx)["__tmp"][0]
            self._meta_dt = remove_row(self._meta_dt, rem_row)
            self._bookmark_df = remove_row(self._bookmark_df, rem_row)
            if dt:
                dt.first().set_dt(self._bookmark_df, self._meta_dt)

        self.post_message(Bookmarks.BookmarkRemoved(selected_index=idx))

    @on(CustomTable.CellSelected)
    def handle_select(self, event: CustomTable.CellSelected):
        event.stop()
        sel_row = event.coordinate.row
        index = int(self._meta_dt[sel_row][INDEX_COL][0])
        self.post_message(Bookmarks.BookmarkSelected(selected_index=index))



from dataclasses import dataclass

from textual import on
from textual.message import Message
from textual.reactive import reactive
from textual.widget import Widget
from textual.widgets import SelectionList
from textual.widgets.selection_list import Selection


class ColumnSelector(Widget):

    DEFAULT_CSS = """
    ColumnSelector {
        dock: right;
    }
    """

    BINDINGS = [
        ("escape", "close()", "Close and Apply"),
        ("a", "apply()", "Apply"),
        ("shift+up", "move(True)", "Move up"),
        ("shift+down", "move(False)", "Move Down"),
    ]

    @dataclass
    class ColumnSelectionChanged(Message):
        selected_columns: tuple[str, ...]
        selector: "ColumnSelector"

        @property
        def control(self):
            return self.selector

    available_columns: reactive[tuple[str, ...]] = reactive(tuple())
    selected_columns: reactive[tuple[str, ...]] = reactive(tuple(), init=False)
    display_columns: reactive[tuple[str, ...]] = reactive(tuple(), init=False, bindings=True)

    def __init__(self, *args, title: str | None = None, allow_reorder: bool = True, **kwargs):
        super().__init__(*args, **kwargs)
        self._allow_reorder = allow_reorder
        self._title = title
        self._message: ColumnSelector.ColumnSelectionChanged | None = None

    def action_close(self):
        self.action_apply()
        self.remove()

    def action_apply(self):
        if self._message is not None:
            self.post_message(self._message)
            self._message = None

    def check_action(self, action: str, parameters: tuple[object, ...]) -> bool | None:
        if action == "move":
            if not self.display_columns or not self._allow_reorder:
                return False
            cur_idx = self.query_one(SelectionList).highlighted
            if cur_idx is None:
                return False
            if parameters[0]:
                return cur_idx > 0
            return cur_idx < len(self.display_columns) - 1
        return True

    @on(SelectionList.SelectionHighlighted)
    def _refresh_actions(self):
        self.refresh_bindings()

    def _refresh_options(self):
        sel_list = self.query_one(SelectionList)
        sel_idx = sel_list.highlighted
        if sel_idx is not None:
            sel_val: str | None = sel_list.get_option_at_index(sel_idx).value
        else:
            sel_val = None
        sel_list.clear_options()
        for i, x in enumerate(self.display_columns):
            sel_list.add_option(Selection(x, x, x in self.selected_columns))
            if x == sel_val:
                sel_list.highlighted = i

    def action_move(self, is_up: bool):
        sel_list = self.query_one(SelectionList)
        if (idx := sel_list.highlighted) is None:
            return
        if is_up:
            self.display_columns = (
                self.display_columns[0 : idx - 1]
                + (self.display_columns[idx], self.display_columns[idx - 1])
                + self.display_columns[idx + 1 :]
            )
        else:
            self.display_columns = (
                self.display_columns[0:idx]
                + (self.display_columns[idx + 1], self.display_columns[idx])
                + self.display_columns[idx + 2 :]
            )
        self._refresh_options()

    def watch_available_columns(self):
        new_disp = []
        for x in self.available_columns:
            if x not in self.display_columns:
                new_disp.append(x)
        if new_disp:
            self.display_columns = self.display_columns + tuple(new_disp)
        self.styles.width = max(max([len(x) for x in self.available_columns] + [0]) + 10, 35)

    def watch_display_columns(self):
        self.selected_columns = [x for x in self.display_columns if x in self.selected_columns]
        self._refresh_options()

    def watch_selected_columns(self):
        self._message = ColumnSelector.ColumnSelectionChanged(selected_columns=self.selected_columns, selector=self)

    def on_mount(self):
        (sel := self.query_one(SelectionList)).focus()
        sel.highlighted = 0

    def compose(self):
        sel = SelectionList[int](
            *(Selection(x, x, x in self.selected_columns) for x in self.display_columns),
        )

        sel.border_title = self._title
        yield sel

    @on(SelectionList.SelectedChanged)
    def on_column_selection(self, event: SelectionList.SelectedChanged):
        event.stop()
        sels = event.selection_list.selected
        self.selected_columns = [x for x in self.display_columns if x in sels]
